#include <stdio.h>
#include <stdlib.h>
#include <strings.h>

#include "fmt_chunk.h"

#define BUFFER_SIZE 131072
#define DIRTY_ITERATIONS 50
#define DIRTY_COUNT 20000
#define DIRTY_INCREMENT 2000

int read_chars(FILE *in, char *s, int n)
{
int t,ch;

  for (t=0; t<n; t++)
  {
    ch=getc(in);
    if (ch==EOF) return -1;
    s[t]=ch;
  }
  s[t]=0;

  return 0;
}

int write_long(FILE *out, int n)
{

  putc((n&255),out);
  putc(((n>>8)&255),out);
  putc(((n>>16)&255),out);
  putc(((n>>24)&255),out);

  return 0;
}

int write_word(FILE *out, int n)
{

  putc((n&255),out);
  putc(((n>>8)&255),out);

  return 0;
}

unsigned int read_long(FILE *in)
{
unsigned int t;

  t=getc(in);
  t=(getc(in)<<8)+t;
  t=(getc(in)<<16)+t;
  t=(getc(in)<<24)+t;

  return t;
}

unsigned short int read_word(FILE *in)
{
unsigned short int t;

  t=getc(in);
  t=(getc(in)<<8)+t;

  return t;
}

int parse_header(FILE *in)
{
int length;
char riff_type[5];

  length=read_long(in);
  read_chars(in,riff_type,4);

  printf("RIFF Header\n");
  printf("----------------------------\n");
  printf("          Length: %d\n",length);
  printf("            Type: %s\n",riff_type);
  printf("----------------------------\n");

  return 0;
}

int parse_fmt(FILE *in, struct fmt_chunk_t *fmt_chunk)
{
int length;

  length=read_long(in);
  fmt_chunk->format_type=read_word(in);
  fmt_chunk->channel_numbers=read_word(in);
  fmt_chunk->sample_rate=read_long(in);
  fmt_chunk->bytes_per_second=read_long(in);
  fmt_chunk->bytes_per_sample=read_word(in);
  fmt_chunk->bits_per_sample=read_word(in);

  printf("FMT Chunk\n");
  printf("----------------------------\n");
  printf("          Length: %d\n",length);
  printf("     Format Type: ");
  if (fmt_chunk->format_type==0)
  { printf("Mono\n"); }
    else
  if (fmt_chunk->format_type==1)
  { printf("Stereo\n"); }
    else
  { printf("unknown\n"); }

  printf(" Channel Numbers: %d\n",fmt_chunk->channel_numbers);
  printf("     Sample Rate: %d\n",fmt_chunk->sample_rate);
  printf("Bytes Per Second: %d\n",fmt_chunk->bytes_per_second);
  printf("Bytes Per Sample: ");
  if (fmt_chunk->bytes_per_sample==1)
  { printf("8 bit mono (%d)\n",fmt_chunk->bytes_per_sample); }
    else
  if (fmt_chunk->bytes_per_sample==2)
  { printf("8 bit stereo or 16 bit mono (%d)\n",fmt_chunk->bytes_per_sample); }
    else
  if (fmt_chunk->bytes_per_sample==4)
  { printf("16 bit stereo (%d)\n",fmt_chunk->bytes_per_sample); }

  printf(" Bits Per Sample: %d\n",fmt_chunk->bits_per_sample);
  printf("----------------------------\n");

  return 0;
}

int parse_data(FILE *in, struct fmt_chunk_t *fmt_chunk, int dirty)
{
short int buffer[BUFFER_SIZE];
unsigned int spikes1[DIRTY_COUNT*2+10];
unsigned int spikes2[DIRTY_COUNT*2+10];
unsigned int spikes1_peak[DIRTY_COUNT+10];
unsigned int spikes2_peak[DIRTY_COUNT+10];
unsigned int delta,spike_ptr;
char hill;
int *spike_curr_ptr,*spike_last_ptr;
int *spike_peak_curr_ptr,*spike_peak_last_ptr;
int *temp;
int length,marker,deepest;
int t,ptr,k,dirty_offset,dirty_count,iterations;
int last_dirty_count,last_dirty_offset,dirty_flag;
short int r;
int p;
int r1;
float ratio,tempratio;

  deepest=0;
  iterations=0;
  /* dirty_offset=100; */
  dirty_offset=DIRTY_INCREMENT;
  last_dirty_offset=0;
  last_dirty_count=0;

  length=read_long(in);
  marker=ftell(in);

  spikes1[0]=-1;
  spikes2[0]=-1;
  spike_curr_ptr=spikes1;
  spike_last_ptr=spikes2;
  spike_peak_curr_ptr=spikes1_peak;
  spike_peak_last_ptr=spikes2_peak;

  printf("DATA chunk\n");
  printf("----------------------------\n");
  printf("          Length: %d\n",length);

  printf("Scanning for biggest/smallest peak\n");

  if (fmt_chunk->bits_per_sample==16)
  {
    for (t=0; t<length/2; t++)
    {
      r=getc(in)+(getc(in)<<8);
      p=abs(r);
      if (p>deepest) deepest=p;
    }
  }

  printf("Clean Deepest: %d\n",deepest);

  if (deepest>32760 && dirty==0)
  {
    printf("This song is already normalized.\n");
    return -1;
  }

  if (dirty==1)
  {
    while(1)
    {
      fseek(in,marker,0);

      dirty_count=0;
      spike_ptr=0;
      dirty_flag=0;
      delta=0;
      hill=1;

      /* for (t=0; t<DIRTY_COUNT; t++) spike_curr_ptr[t]=0; */

      for (t=0; t<length/2; t++)
      {
        r=getc(in)+(getc(in)<<8);
        if (r<0 && hill==1)
        {
          hill=0;

          /* if (dirty_flag==-1) */
          if (dirty_flag!=0)
          {
            spike_curr_ptr[spike_ptr]=delta;
            spike_ptr++;
            spike_curr_ptr[spike_ptr]=t;
            spike_ptr++;
            dirty_flag=0;
          }

          delta=t;
        }
          else
        if (r>=0 && hill==0)
        {
          hill=1;

          /* if (dirty_flag==-1) */
          if (dirty_flag!=0)
          {
            spike_curr_ptr[spike_ptr]=delta;
            spike_ptr++;
            spike_curr_ptr[spike_ptr]=t;
            spike_ptr++;
            dirty_flag=0;
          }

          delta=t;
        }

        p=abs(r);
        if (p>deepest-dirty_offset && (dirty_flag==0 || dirty_flag==-1))
        {
          dirty_count++;

          if (dirty_flag==0)
          { spike_peak_curr_ptr[spike_ptr>>1]=p; }
            else
          if (p>spike_peak_curr_ptr[spike_ptr>>1]) 
          { spike_peak_curr_ptr[spike_ptr>>1]=p; }

          dirty_flag=1;
        }
          else
        if (p<deepest-dirty_offset && dirty_flag==1)
        {
          dirty_flag=-1;
        }
          else
        if (p>deepest-dirty_offset && dirty_flag==1)
        {
          if (p>spike_peak_curr_ptr[spike_ptr>>1]) 
          { spike_peak_curr_ptr[spike_ptr>>1]=p; }
        }

        if (dirty_count>DIRTY_COUNT) break;
      }

      printf("Dirty peak counts: %d\n",dirty_count);

      spike_curr_ptr[spike_ptr]=-1;
      spike_curr_ptr[spike_ptr+1]=-1;

      temp=spike_peak_curr_ptr;
      spike_peak_curr_ptr=spike_peak_last_ptr;
      spike_peak_last_ptr=temp;

      temp=spike_curr_ptr;
      spike_curr_ptr=spike_last_ptr;
      spike_last_ptr=temp;

      if (dirty_count>DIRTY_COUNT) break;

      last_dirty_count=dirty_count;
      last_dirty_offset=dirty_offset;
      dirty_offset=dirty_offset+DIRTY_INCREMENT;

      iterations++;
      if (iterations>=DIRTY_ITERATIONS) break;
    }

    deepest=deepest-last_dirty_offset;
    printf("Dirty deepest: %d\n",deepest-last_dirty_offset);

    if (deepest>32760)
    {
      printf("This song is already normalized.\n");
      return -1;
    }
  }

  fseek(in,marker,0);

  spike_ptr=0;
  hill=0;

  ratio=32767/(float)deepest;
  printf("Ratio: %.5f\n",ratio);
  printf("Rewriting wave\n");

  /* Write Data Chunk */

#ifdef DEBUG
  for (t=0; t<42; t++)
  {
    if (t%2==0) printf("(");
    printf("%d ",spike_curr_ptr[t]);
    if (t%2==1) printf(")");
  }
  printf("\n");
#endif

  if (fmt_chunk->bits_per_sample==16)
  {
    ptr=0;

    for (t=0; t<length/2; t++)
    {
      r=getc(in)+(getc(in)<<8);

      if (spike_curr_ptr[spike_ptr]==t)
      {
        if (hill==1)
        {
          hill=0;
#ifdef DEBUG
          printf("Hill end %d\n\n",t);
#endif
          spike_ptr++;

          if (spike_curr_ptr[spike_ptr]==t)
          {
            hill=1;
            tempratio=32767/(float)spike_peak_curr_ptr[spike_ptr>>1];
            spike_ptr++;
#ifdef DEBUG
            printf("Hill start %d new ratio %.5f with a peak %d\n",t,tempratio,spike_peak_curr_ptr[spike_ptr>>1]);
#endif
          }
        }
          else
        {
#ifdef DEBUG
if (spike_ptr%2!=0) printf("Warning: Spike's out of sync\n");
#endif
          hill=1;
          tempratio=32767/(float)spike_peak_curr_ptr[spike_ptr>>1];
          spike_ptr++;
#ifdef DEBUG
          printf("Hill start %d new ratio %.5f with a peak %d\n",t,tempratio,spike_peak_curr_ptr[spike_ptr>>1]);
#endif
        }
      }

      if (hill==0)
      { r1=(int)((float)r*ratio); }
        else
      { r1=(int)((float)r*tempratio); }

      if (r1>32767)
      {
        printf("Clip warning at: %d   volume: %d  r=%d  ratio=%.5f  hill=%d\n",t,r1,r,ratio,hill);
        r1=32767;
      }
        else
      if (r1<-32767)
      {
        printf("Clip warning at: %d   volume: %d  r=%d  ratio=%.5f  hill=%d\n",t,r1,r,ratio,hill);
        r1=-32767;
      }

      r=r1;

      buffer[ptr++]=r;
      if (ptr>=BUFFER_SIZE)
      {
        fseek(in,-(ptr*2), SEEK_CUR);
        for (k=0; k<ptr; k++)
        { write_word(in,buffer[k]); }
        ptr=0;
      }
/*
      fseek(in,-2, SEEK_CUR);
      write_word(in,r);
*/
    }

#ifdef DEBUG
    printf("next hill starts at: %d\n",spike_curr_ptr[spike_ptr]);
#endif

    if (ptr!=0)
    {
      fseek(in,-(ptr*2), SEEK_CUR);
      for (k=0; k<ptr; k++)
      { write_word(in,buffer[k]); }
    }
  }
    else
  if (fmt_chunk->bits_per_sample==8)
  {
    printf("8 bit wav's not supported\n");
  }

  return 0;
}

int normalize(char *inname, int dirty)
{
FILE *in;
char chunk_name[5];
struct fmt_chunk_t fmt_chunk;

  in=fopen(inname,"rb+");
  if (in==0)
  {
    printf("Couldn't open file for reading: %s\n",inname);
    return -1;
  }

  while(1)
  {
    if (read_chars(in,chunk_name,4)!=0) break;

    if (strcasecmp(chunk_name,"RIFF")==0)
    { parse_header(in); }
      else
    if (strcasecmp(chunk_name,"fmt ")==0)
    { parse_fmt(in,&fmt_chunk); }
      else
    if (strcasecmp(chunk_name,"data")==0)
    { if (parse_data(in,&fmt_chunk,dirty)==-1) break; }
      else
    {
      printf("Unknown chunk: '%s'\n",chunk_name);
      return -1;
    }
  }

  fclose(in);

  return 0;
}

#ifdef STANDALONE
int main()
{
  normalize("/storage/test.wav",1);
  return 0;
}
#endif

